// Christian Elzey

// The code for I2C communication with the DS3231 RTC borrows hevily
// from Bruce E. Hall`s tutorial on adding a 1307 RTC to an AVR project
// http://w8bh.net/avr/AvrDS1307.pdf
 
                                          


#include <avr/io.h>                        /* Defines pins, ports, etc */
#include <util/delay.h>                     /* Functions to waste time */
#include <avr/interrupt.h>					
#include <avr/pgmspace.h>

   // function prototypes
   void shiftWrite(uint8_t data);
   void latchUp();
   void set_time();
   void get_time();
   uint8_t decToBcd(uint8_t val);
   uint8_t bcdToDec(uint8_t val);
   void setupDisplay();

uint8_t digit[8]; //stores state of each tube
uint8_t bits32[8]; //this is whats sent to the driver IC
uint8_t charDisplay[11]; //stores character info for 7 seg displays
uint8_t hours = 12;
uint8_t minutes = 34;
uint8_t month = 56;
uint8_t date = 78;
uint8_t year = 18;
uint8_t day = 1;
uint8_t seconds = 0;

uint8_t ROW = 0;

#define F_CPU 8000000 // run CPU at 16 MHz
#define LED 5 // Boarduino LED on PB5
#define ClearBit(x,y) x &= ~_BV(y) // equivalent to cbi(x,y)
#define SetBit(x,y) x |= _BV(y) // equivalent to sbi(x,y)


#include <avr/io.h> // deal with port registers
#include <util/delay.h> // used for _delay_ms function
#include <string.h> // string manipulation routines
#include <stdlib.h>


typedef uint8_t byte; // I just like byte & sbyte better
typedef int8_t sbyte;

void InitAVR();
#define F_SCL 100000L // I2C clock speed 100 KHz
#define READ 1
#define TW_START 0xA4 // send start condition (TWINT,TWSTA,TWEN)
#define TW_STOP 0x94 // send stop condition (TWINT,TWSTO,TWEN)
#define TW_ACK 0xC4 // return ACK to slave
#define TW_NACK 0x84 // don't return ACK to slave
#define TW_SEND 0x84 // send data (TWINT,TWEN)
#define TW_READY (TWCR & 0x80) // ready when TWINT returns to logic 1.
#define TW_STATUS (TWSR & 0xF8) // returns value of status register
#define I2C_Stop() TWCR = TW_STOP // inline macro for stop condition

#define DS1307 0xD0 // I2C bus address of DS1307 RTC
#define SECONDS_REGISTER 0x00
#define MINUTES_REGISTER 0x01
#define HOURS_REGISTER 0x02
#define DAYOFWK_REGISTER 0x03
#define DAYS_REGISTER 0x04
#define MONTHS_REGISTER 0x05
#define YEARS_REGISTER 0x06
#define CONTROL_REGISTER 0x07
#define RAM_BEGIN 0x08
#define RAM_END 0x3F

void I2C_Init();
byte I2C_Detect(byte addr);
void I2C_Start (byte slaveAddr);
byte I2C_Write (byte data);
byte I2C_ReadACK ();
byte I2C_ReadNACK ();
void I2C_WriteRegister(byte busAddr, byte deviceRegister, byte data);
byte I2C_ReadRegister(byte busAddr, byte deviceRegister);
void DS1307_GetTime();
void DS1307_GetDate();
void SetDateTime();
void checkButtons();
void LEDon();
void LEDoff();



void I2C_Init()
// at 16 MHz, the SCL frequency will be 16/(16+2(TWBR)), assuming prescalar of 0.
// so for 100KHz SCL, TWBR = ((F_CPU/F_SCL)-16)/2 = ((16/0.1)-16)/2 = 144/2 = 72.
{
 TWSR = 0; // set prescalar to zero
 TWBR = ((F_CPU/F_SCL)-16)/2; // set SCL frequency in TWI bit register
}


byte I2C_Detect(byte addr)
// look for device at specified address; return 1=found, 0=not found
{
 TWCR = TW_START; // send start condition
 while (!TW_READY); // wait
 TWDR = addr; // load device's bus address
 TWCR = TW_SEND; // and send it
 while (!TW_READY); // wait
 return (TW_STATUS==0x18); // return 1 if found; 0 otherwise
}


void I2C_Start (byte slaveAddr)
{
 I2C_Detect(slaveAddr);
}

byte I2C_Write (byte data) // sends a data byte to slave
{
 TWDR = data; // load data to be sent
 TWCR = TW_SEND; // and send it
 while (!TW_READY); // wait
 return (TW_STATUS!=0x28);
}

byte I2C_ReadACK () // reads a data byte from slave
{
 TWCR = TW_ACK; // ack = will read more data
 while (!TW_READY); // wait
 return TWDR;
 //return (TW_STATUS!=0x28);
}


byte I2C_ReadNACK () // reads a data byte from slave
{
 TWCR = TW_NACK; // nack = not reading more data
 while (!TW_READY); // wait
 return TWDR;
 //return (TW_STATUS!=0x28);
}


void I2C_WriteRegister(byte busAddr, byte deviceRegister, byte data)
{
 I2C_Start(busAddr); // send bus address
 I2C_Write(deviceRegister); // first byte = device register address
 I2C_Write(data); // second byte = data for device register
 I2C_Stop();
}

void LEDon(){
PORTC |= 0b00001000;
}
void LEDoff(){
PORTC &= 0b11110111;
}

byte I2C_ReadRegister(byte busAddr, byte deviceRegister)
{
 byte data = 0;
 I2C_Start(busAddr); // send device address
 I2C_Write(deviceRegister); // set register pointer
 I2C_Start(busAddr+READ); // restart as a read operation
 data = I2C_ReadNACK(); // read the register data
 I2C_Stop(); // stop
 return data;
}

void DS1307_GetTime()
// returns hours, minutes, and seconds in BCD format
{
 hours = bcdToDec(I2C_ReadRegister(DS1307,HOURS_REGISTER));
 minutes = bcdToDec(I2C_ReadRegister(DS1307,MINUTES_REGISTER));
 seconds = bcdToDec(I2C_ReadRegister(DS1307,SECONDS_REGISTER));
/* if (hours & 0x40) // 12hr mode:
 hours &= 0x1F; // use bottom 5 bits (pm bit = temp & 0x20)
 else hours &= 0x3F; // 24hr mode: use bottom 6 bits*/
}

void DS1307_GetDate()
// returns months, days, and years in BCD format
{
 month = bcdToDec(I2C_ReadRegister(DS1307,MONTHS_REGISTER));
 date = bcdToDec(I2C_ReadRegister(DS1307,DAYS_REGISTER));
 year = bcdToDec(I2C_ReadRegister(DS1307,YEARS_REGISTER));
}

void SetDateTime()
// simple, hard-coded way to set the date.
{
 I2C_WriteRegister(DS1307,MONTHS_REGISTER, decToBcd(month));
 I2C_WriteRegister(DS1307,DAYS_REGISTER, decToBcd(date));
 I2C_WriteRegister(DS1307,YEARS_REGISTER, decToBcd(year));
 I2C_WriteRegister(DS1307,HOURS_REGISTER, decToBcd(hours)); // add 0x40 for PM
 I2C_WriteRegister(DS1307,MINUTES_REGISTER, decToBcd(minutes));
 I2C_WriteRegister(DS1307,SECONDS_REGISTER, decToBcd(seconds));
}



int main(void) {
  // -------- Inits --------- //
  DDRD |= 0b00101110;                 // all chip pins are output
  DDRC |= 0b00001000;                 //       LED is only output on port C  
  PORTD &= ~(1 << 5);                 // write the Enable pin on VFD driver low. Could be PWM'd. 
  PORTC = 0b11110111;                 // enable pull up resistors on port C 
       
// Set up variables

  charDisplay[0] = 0b11111100;
  charDisplay[1] = 0b01100000;
  charDisplay[2] = 0b11011010;
  charDisplay[3] = 0b11110010;
  charDisplay[4] = 0b01100110;
  charDisplay[5] = 0b10110110;
  charDisplay[6] = 0b10111110;
  charDisplay[7] = 0b11100000;
  charDisplay[8] = 0b11111110;
  charDisplay[9] = 0b11110110;
  charDisplay[10] = 0b00000000;
                       
   setupDisplay();
  // ------ Event loop ------ //
  
  cli();
  TCCR0A = 0;// set entire TCCR2A register to 0
  TCCR0B = 0;// same for TCCR2B
  TCNT0  = 0;//initialize counter value to 0
  // set compare match register for 2khz increments
  OCR0A = 124;// = (16*10^6) / (2000*64) - 1 (must be <256)
  // turn on CTC mode
  TCCR0A |= (1 << WGM01);
  // Set CS01 and CS00 bits for 64 prescaler
  TCCR0B |= (1 << CS01) | (1 << CS00);   
  // enable timer compare interrupt
  TIMSK0 |= (1 << OCIE0A);
  sei();
  
  while (1) {//LOOP
     _delay_ms(300);
     DS1307_GetTime();  
     DS1307_GetDate();
     setupDisplay();
     checkButtons(); 

  }                                                  /* End event loop */
  return 0;                            /* This line is never reached */
}//end main


void checkButtons(){ // This is the function for setting the time. Uses polling. Super simple. 

if((PINC & 0b00000001)== 0) // if hour / the month button pressed...
	{ 
		if((PINC & 0b00000010) == 0)	// if date set button held
			{
				month = month + 1;
					if(month > 12){ month = 1;}
			}
		else
			{
			hours ++;
			if(hours > 23){hours = 0;}
			}
		SetDateTime();	 // save changes to RTC chip
	}
if((PINC & 0b00000100) == 0) // if minute / day of month button pressed
	{
		if((PINC & 0b00000010) == 0)
			{
				date ++;
				if(date > 31){date = 1;}
			}	
		else{
			minutes ++;
			if(minutes > 59){ minutes = 0;}
			}
			SetDateTime(); // save changes to RTC chip
	}
}// end checkButtons




void latchUp(){
   //Move them to output latch at one
      PORTD |= (1<<PD3); // pulse clock
      PORTD &= ~(1<<PD3);
}

void shiftWrite(uint8_t data) // hard coded to the "data" pin on MAX6934
{
   //Send each 8 bits serially
   //Order is MSB first
   for(uint8_t i=0;i<8;i++)
   {
      //Output the data on DS line according to the
      //Value of MSB
      if(data & 0b00000001)
      {
         //LSB is 1 so output high
          PORTD |= (1<<PD1);
      }
      else
      {
         //LSB is 0 so output low
            PORTD &= ~(1<<PD1);
      }
      PORTD |= (1<<PD2); // pulse clock
      PORTD &= ~(1<<PD2);
      
      data=data>>1;  //Now bring next bit at LSB position
   }
}

uint8_t decToBcd(uint8_t val)
{
  return ( (val/10*16) + (val%10) );
}

uint8_t bcdToDec(uint8_t val)
{
  return ( (val/16*10) + (val%16) );
}

void formatTime(){
uint8_t fakeHours = hours;

if(fakeHours  > 12)
	
			{ fakeHours = fakeHours - 12;
			LEDon();}
	else{ LEDoff();}
if(fakeHours == 0){ fakeHours = 12;}
digit[1] = fakeHours % 10;
digit[0] = (fakeHours / 10) % 10;
digit[3] = minutes % 10;
digit[2] = (minutes / 10 ) % 10;
digit[5] = month % 10;
digit[4] = (month / 10) % 10;
digit[7] = date % 10;
digit[6] = (date / 10 ) % 10;
}

void writeTubes(){
if(ROW){
shiftWrite(bits32[4]);
shiftWrite(bits32[5]);
shiftWrite(bits32[6]);
shiftWrite(bits32[7]);
latchUp();
ROW = 0;
}

else{
shiftWrite(bits32[0]);
shiftWrite(bits32[1]);
shiftWrite(bits32[2]);
shiftWrite(bits32[3]);
latchUp();
ROW = 1;
}
}

ISR(TIMER0_COMPA_vect){ // The internal timer calls this routine at 2KHz. 
writeTubes();
}

void setupDisplay(){

formatTime();
  
bits32[3] = 0;
bits32[2] = 0;
bits32[1] = 0;
bits32[0] = 0;
bits32[4] = 0;
bits32[5] = 0;
bits32[6] = 0;
bits32[7] = 0;

bits32[7] = bits32[7] | (charDisplay[digit[4]] );
bits32[7] = bits32[7] | (charDisplay[digit[5]] >> 7);
bits32[6] = bits32[6] | (charDisplay[digit[5]] << 1);
bits32[6] = bits32[6] | (charDisplay[digit[6]] >> 6);
bits32[5] = bits32[5] | (charDisplay[digit[6]] << 2);
bits32[5] = bits32[5] | (charDisplay[digit[7]] >> 5);
bits32[4] = bits32[4] | (charDisplay[digit[7]] << 3);
bits32[4] = bits32[4] | 0b00000100; 
bits32[3] = bits32[3] | (charDisplay[digit[0]] );
bits32[3] = bits32[3] | (charDisplay[digit[1]] >> 7);
bits32[2] = bits32[2] | (charDisplay[digit[1]] << 1);
bits32[2] = bits32[2] | (charDisplay[digit[2]] >> 6);
bits32[1] = bits32[1] | (charDisplay[digit[2]] << 2);
bits32[1] = bits32[1] | (charDisplay[digit[3]] >> 5);
bits32[0] = bits32[0] | (charDisplay[digit[3]] << 3);
bits32[0] = bits32[0] | 0b00001000;
}